require 'securerandom'

module Vx
  module Common
    module AMQP
      module Consumer::Subscribe

        def subscribe
          start do |x, q|
            subscription_loop q
          end
        end

        def start
          rs = nil
          instrumentation = {
            consumer_id: consumer_id,
            consumer:    consumer_name
          }
          session.open instrumentation
          session.with_channel do
            x = declare_exchange
            q = declare_queue

            instrumentation.merge!(
              exchange:    x.name,
              queue:       q.name,
            )
            instrument("start.consumer.amqp", instrumentation)

            q.bind(x, bind_options)
            rs = yield(x, q) if block_given?

            instrument("shutdown.consumer.amqp", instrumentation)
          end
          rs
        end

        def pop(q)
          unpacked = nil
          delivery_info, properties, payload = q.pop(ack: ack)

          if payload
            unpacked = deserialize_message properties, payload
          end

          [unpacked, delivery_info, properties]
        end

        private

          def subscription_loop(q)
            loop do
              break if shutdown?

              delivery_info, properties, payload = q.pop(ack: ack)

              if payload
                payload = deserialize_message properties, payload
                result  = nil
                uuid    = SecureRandom.uuid

                instrumentation = {
                  consumer_id:   consumer_id,
                  consumer:      consumer_name,
                  payload:       payload,
                  properties:    properties,
                }

                with_sub_middlewares instrumentation do
                  instrument("start_processing.consumer.amqp", instrumentation)
                  instrument("process.consumer.amqp", instrumentation) do
                    result  = run_instance delivery_info, properties, payload, uuid
                  end
                end

                break if result == :shutdown
              else
                sleep config.pool_timeout
              end
            end
          end

          def run_instance(delivery_info, properties, payload, uuid)
            new.tap do |inst|
              inst.properties    = properties
              inst.delivery_info = delivery_info
              inst.request_id    = uuid
            end.perform payload
          end

          def with_sub_middlewares(env, &block)
            Common::AMQP.config.builders[:sub].to_app(block).call(env)
          end

          def deserialize_message(properties, payload)
            Common::AMQP::Formatter.unpack(
              properties[:content_type],
              model,
              payload
            )
          end

      end
    end
  end
end
